Add LoadPath "basis".


(** * Permutation over lists, and finite multisets. *)

Set Implicit Arguments. 

Require Import decidable_set.
Require Import List.
Require Import more_list.
Require Import equiv_list.
Require Import Multiset.
Require Import Arith.
Require Import Setoid.

Module Type Permut.

  Declare Module DS : decidable_set.S.

  Definition elt := DS.A.
  Definition eq_elt_dec := DS.eq_A_dec.

  Fixpoint list_to_multiset (l : list elt) {struct l} : multiset elt :=
    match l with
      | nil => EmptyBag elt
      | h :: tl =>
	munion (SingletonBag _ eq_elt_dec h) (list_to_multiset tl)
    end.

  Definition list_permut (l1 l2:list elt) : Prop :=
    meq (list_to_multiset l1) (list_to_multiset l2).

 Axiom list_permut_refl :
    forall (l : list elt), list_permut l l.

  Axiom list_permut_sym :
    forall l1 l2 : list elt, list_permut l1 l2 -> list_permut l2 l1.

  Axiom list_permut_trans :
    forall l1 l2 l3 : list elt, list_permut l1 l2 -> list_permut l2 l3 -> list_permut l1 l3.

  Axiom list_permut_nil :
    forall l, list_permut l nil -> l = nil.

  Axiom in_permut_in :
    forall l1 l2 e, list_permut l1 l2 -> (In e l1 <-> In e l2).

  Axiom cons_permut_in :
    forall l1 l2 e, list_permut (e :: l1) l2 -> In e l2.

  Axiom permut_cons :
    forall e l1 l2, list_permut l1 l2 <-> list_permut (e :: l1) (e :: l2).

  Axiom permut_add_inside :
    forall e l1 l2 l3 l4, 
      list_permut (l1 ++ l2) (l3 ++ l4) <->
      list_permut (l1 ++ e :: l2) (l3 ++ e :: l4).

  Axiom permut_cons_inside :
    forall a l l1 l2, 
      list_permut l (l1 ++ l2) <->
      list_permut (a :: l) (l1 ++ a :: l2).

Axiom permut_app1 :
  forall l l1 l2, list_permut l1 l2 <-> list_permut (l ++ l1) (l ++ l2).

Axiom permut_app2 :
  forall l l1 l2, list_permut l1 l2 <-> list_permut (l1 ++ l) (l2 ++ l).

Axiom list_permut_app_app :
 forall l1 l2, list_permut (l1 ++ l2) (l2 ++ l1).

Axiom list_permut_length :
 forall l1 l2, list_permut l1 l2 -> length l1 = length l2.

Axiom list_permut_size :
  forall size l1 l2, list_permut l1 l2 -> list_size size l1 = list_size size l2.

Axiom list_permut_size_equiv :
  forall size ll l1 l2, list_permut l1 (map (fun st =>fst st) ll) ->
       list_permut l2 (map (fun st => snd st) ll) -> 
       (forall t1 t2, In (t1,t2) ll -> size t1 = size t2) ->
       list_size size l1 = list_size size l2.

  Axiom list_permut_map :
  forall f l1 l2, list_permut l1 l2 -> list_permut (map f l1) (map f l2).

Axiom list_permut_length_1:
 forall a b, list_permut (a :: nil) (b :: nil)  -> a = b.

Axiom list_permut_length_2 :
 forall a1 b1 a2 b2, list_permut (a1 :: b1 :: nil) (a2 :: b2 :: nil) ->
 (a1=a2 /\ b1=b2) \/ (a1=b2 /\ a2=b1).

Axiom ac_syntactic :
 forall (l1 l2 l3 l4 : list elt),
 list_permut (l2 ++ l1) (l4 ++ l3) ->
 (exists u1, exists u2, exists u3, exists u4, 
 list_permut l1 (u1 ++ u2) /\
 list_permut l2 (u3 ++ u4) /\
 list_permut l3 (u1 ++ u3) /\
 list_permut l4 (u2 ++ u4)).

Axiom list_permut_dec : 
  forall (l1 l2 : list elt), {list_permut l1 l2}+{~list_permut l1 l2}.

Axiom permut_filter :
 forall P (P_dec : forall a, {P a}+{~P a}) (l : list elt),
  list_permut l ((filter (fun a => if P_dec a then true else false) l)++
                         (filter (fun a => if P_dec a then false else true) l)).

Axiom permut_list_exists :
  forall f l l', list_permut l l' -> list_exists f l = list_exists f l'.

Axiom trans_equiv_list :
  forall (R : relation elt), forall ll ll',
  (forall e1, In e1 (map (fun st => fst st) ll) ->
        forall e2 e3, R e1 e2 -> R e2 e3 -> R e1 e3) ->
  (forall e1 e2, In (e1,e2) ll -> R e1 e2) -> 
  (forall e1 e2, In (e1,e2) ll' -> R e1 e2) -> 
  list_permut (map (fun st => snd st) ll) (map (fun st => fst st) ll') -> 
     exists ll'', (forall e1 e2, In (e1,e2) ll'' -> R e1 e2) /\
             list_permut (map (fun st => fst st) ll)
                                 (map (fun st => fst st) ll'') /\
             list_permut (map (fun st => snd st) ll') 
                                 (map (fun st => snd st) ll'').

End Permut.

(** ** Definition of permutation over lists. *)
Module Make (DS1 : decidable_set.S) <: Permut with Module DS:= DS1.

  Module DS := DS1.
  Import DS1.

  Definition elt := DS.A.
  Definition eq_elt_dec : forall t1 t2 : elt, {t1 = t2} + {t1 <> t2} := DS.eq_A_dec.

  Function list_to_multiset (l : list elt) {struct l} : multiset elt :=
    match l with
      | nil => EmptyBag elt
      | h :: tl =>
	munion (SingletonBag _ eq_elt_dec h) (list_to_multiset tl)
    end.

  Definition list_permut (l1 l2:list elt) : Prop :=
    meq (list_to_multiset l1) (list_to_multiset l2).

  (** Properties over the multiplicity. *)
  Lemma multiplicity_app :
    forall (l1 l2:list elt) (t : elt),
      multiplicity (list_to_multiset (l1 ++ l2)) t =
      multiplicity (list_to_multiset l1) t + multiplicity (list_to_multiset l2) t.
  Proof.
    induction l1; intros; trivial.
    simpl; intros; rewrite (IHl1 l2 t); auto with arith.
  Qed.

  Lemma out_mult_O :
    forall (t : elt) (l:list elt), ~ In t l <-> multiplicity (list_to_multiset l) t = 0.
  Proof.
    intros t l; induction l as [ | a l].
    simpl; intuition.
    split; [intros t_not_in_a_l | intros mult_eq_0].
    simpl; destruct (eq_elt_dec a t) as [a_eq_t | _].
    subst t; absurd (In a (a :: l)); trivial; left; trivial.
    simpl; setoid_rewrite <- IHl; intro; apply t_not_in_a_l; right; trivial.
    simpl in mult_eq_0; destruct (eq_elt_dec a t) as [ a_eq_t | a_diff_t].
    discriminate.
    intros [t_eq_a | t_in_l].
    absurd (a = t); trivial.
    generalize t_in_l; setoid_rewrite IHl; trivial.
  Qed.

  Lemma in_mult_S :
    forall (t : elt) (l : list elt), In t l <-> multiplicity (list_to_multiset l) t >= 1.
  Proof.
    intros t l; split; induction l as [ | a l].
    (* -> case *)
    contradiction.
    intros [a_eq_t | t_in_l]; simpl; destruct (eq_elt_dec a t) as [_ | t_diff_a].
    auto with arith.
    absurd (a = t); trivial.
    auto with arith.
    apply IHl; trivial.
    simpl; unfold ge; intro one_le_0; inversion one_le_0.
    (* <- case *)
    simpl; destruct (eq_elt_dec a t); auto with arith.
  Qed.


  (** ** Permutation is a equivalence relation. 
      Reflexivity. *)
  Theorem list_permut_refl :
    forall (l : list elt), list_permut l l.
  Proof.
    unfold list_permut, meq; intros; trivial.
  Qed.

  (** Symetry. *)
  Theorem list_permut_sym :
    forall l1 l2 : list elt, list_permut l1 l2 -> list_permut l2 l1.
  Proof.
    unfold list_permut, meq; intros; apply sym_eq; trivial.
  Qed.

  Hint Immediate list_permut_refl.
  Hint Resolve list_permut_sym.

  (** Transitivity. *)
  Theorem list_permut_trans :
    forall l1 l2 l3 : list elt, list_permut l1 l2 -> list_permut l2 l3 -> list_permut l1 l3.
  Proof.
    unfold list_permut, meq; intros;
      apply trans_eq with (multiplicity (list_to_multiset l2) a); trivial.
  Qed.

  Add Relation (list elt) list_permut 
  reflexivity proved by list_permut_refl
    symmetry proved by list_permut_sym
      transitivity proved by list_permut_trans as LP.

  (** Permutation of an empty list. *)
  Lemma list_permut_nil :
    forall l, list_permut l nil -> l = nil.
  Proof.
    intros l P; destruct l as [ | e l ]; trivial.
    unfold list_permut, meq in P; assert (H := P e); simpl in H.
    destruct (eq_elt_dec e e) as [ _ | e_diff_e]; [discriminate | absurd (e = e); trivial].
  Qed.


  (** ** Compatibility Properties. 
      Permutation is compatible with In. *)
  Lemma in_permut_in :
    forall l1 l2 e, list_permut l1 l2 -> (In e l1 <-> In e l2).
  Proof.
    unfold list_permut, meq; intros l1 l2 e P; split; intro H.
    setoid_rewrite in_mult_S; rewrite <- P; setoid_rewrite <- in_mult_S; trivial.
    setoid_rewrite in_mult_S; rewrite P; setoid_rewrite <- in_mult_S; trivial.
  Qed.

  Add Morphism (In (A :=elt)) : in_morph.
  Proof.
    intros; apply in_permut_in; trivial.
  Qed.

  Lemma cons_permut_in :
    forall l1 l2 e, list_permut (e :: l1) l2 -> In e l2.
  Proof.
    intros l1 l2 e P; setoid_rewrite <- P; left; trivial.
  Qed.

  (** Permutation is compatible with addition and removal of common elements *)
  Lemma permut_cons :
    forall e l1 l2, list_permut l1 l2 <-> list_permut (e :: l1) (e :: l2).
  Proof.
    intros e l1 l2; unfold list_permut, meq; simpl; split; intros H a.
    rewrite (H a); trivial.
    apply (plus_reg_l _ _ _ (H a)).
  Qed.

  Add Morphism (List.cons (A:=elt)) : add_elt_morph.
  Proof.
    intros e l1 l2 P; setoid_rewrite <- (permut_cons e l1 l2); trivial.
  Qed.
  
  Lemma permut_add_inside :
    forall e l1 l2 l3 l4, 
      list_permut (l1 ++ l2) (l3 ++ l4) <->
      list_permut (l1 ++ e :: l2) (l3 ++ e :: l4).
  Proof.
    intros e l1 l2 l3 l4; unfold list_permut, meq; simpl; split; intros H a.
    (* -> case *)
    do 2 rewrite multiplicity_app; simpl;
      rewrite plus_comm; rewrite <- plus_assoc; 
	rewrite (plus_comm (multiplicity (list_to_multiset l2) a));
	  rewrite <- multiplicity_app;
	    rewrite H;
	      rewrite multiplicity_app;
		rewrite (plus_comm (multiplicity (list_to_multiset l3) a));
		  rewrite plus_assoc; rewrite plus_comm; trivial.
    (* <- case *)
    apply plus_reg_l with (if eq_elt_dec e a then 1 else 0);
      do 2 rewrite multiplicity_app; simpl;
	rewrite plus_comm;
	  rewrite <- plus_assoc;
	    rewrite (plus_comm (multiplicity (list_to_multiset l2) a)).
    assert (Ha := H a); rewrite multiplicity_app in Ha; simpl in Ha; rewrite Ha; clear Ha.
    rewrite multiplicity_app; simpl;
      rewrite plus_comm;
	rewrite <- plus_assoc;
	  rewrite (plus_comm (multiplicity (list_to_multiset l4) a)); trivial.
  Qed.

  Lemma permut_cons_inside :
    forall a l l1 l2, 
      list_permut l (l1 ++ l2) <->
      list_permut (a :: l) (l1 ++ a :: l2).
  Proof.
    intros a l l1 l2; apply (permut_add_inside a nil l).
  Qed.

(** Permutation is compatible with append. *)
Lemma permut_app1 :
  forall l l1 l2, list_permut l1 l2 <-> list_permut (l ++ l1) (l ++ l2).
Proof.
intros l l1 l2; unfold list_permut, meq; simpl; split.
(* -> case *)
intros H a; 
rewrite (multiplicity_app l l1); rewrite (multiplicity_app l l2); rewrite (H a); 
trivial.
(* <- case *)
intros H a; generalize (H a);
rewrite (multiplicity_app l l1);
rewrite (multiplicity_app l l2); 
intros; apply plus_reg_l with (multiplicity (list_to_multiset l) a); trivial.
Qed.

Lemma permut_app2 :
  forall l l1 l2, list_permut l1 l2 <-> list_permut (l1 ++ l) (l2 ++ l).
Proof.
intros l l1 l2; unfold list_permut, meq; simpl; split.
(* -> case *)
intros H a; 
rewrite (multiplicity_app l1 l); rewrite (multiplicity_app l2 l); rewrite (H a); 
trivial.
(* <- case *)
intros H a; assert (Ha := H a);
rewrite (multiplicity_app l1 l) in Ha;
rewrite (multiplicity_app l2 l) in Ha; 
intros; apply plus_reg_l with (multiplicity (list_to_multiset l) a); 
rewrite plus_comm; rewrite Ha; rewrite plus_comm; 
trivial.
Qed.

Add Morphism (List.app (A:=elt)) : app_morph.
Proof.
intros.
apply list_permut_trans with (x1 ++ x3).
setoid_rewrite <- permut_app1; trivial.
setoid_rewrite <- permut_app2; trivial.
Qed.

Lemma list_permut_app_app :
 forall l1 l2, list_permut (l1 ++ l2) (l2 ++ l1).
Proof.
intros l1 l2;
unfold list_permut, meq; 
intro a; rewrite multiplicity_app; rewrite multiplicity_app;
apply plus_comm.
Qed.

(** Permutation is compatible with length. *)
Lemma list_permut_length :
 forall l1 l2, list_permut l1 l2 -> length l1 = length l2.
Proof.
induction l1 as [ | a l1 ]; intros l2 P.
rewrite (list_permut_nil (list_permut_sym P)); auto.
destruct (In_split _ _ (cons_permut_in P)) as [l2' [l2'' H]]; subst l2.
setoid_rewrite <- (permut_add_inside a nil l1) in P; simpl in P.
simpl; rewrite (IHl1 (l2' ++ l2'') P); do 2 rewrite length_app; simpl.
rewrite (plus_comm (length l2') (S (length l2''))); simpl; rewrite plus_comm; trivial.
Qed.


Add Morphism (length (A:=elt)) : length_morph.
Proof.
apply list_permut_length.
Qed.

(** Permutation is compatible with size. *)
Lemma list_permut_size :
  forall size l1 l2, list_permut l1 l2 -> list_size size l1 = list_size size l2.
Proof.
intro size; induction l1 as [ | a l1 ]; intros l2 P. 
rewrite (list_permut_nil (list_permut_sym P)); auto.
destruct (In_split _ _ (cons_permut_in P)) as [l2' [l2'' H]]; subst l2.
setoid_rewrite <- (permut_add_inside a nil l1) in P; simpl in P.
simpl; rewrite (IHl1 (l2' ++ l2'') P); do 2 rewrite list_size_app; simpl.
rewrite plus_comm; rewrite <- plus_assoc; rewrite (plus_comm (size a)); trivial.
Qed.

Lemma list_permut_size_equiv :
  forall size ll l1 l2, list_permut l1 (map (fun st =>fst st) ll) ->
       list_permut l2 (map (fun st => snd st) ll) -> 
       (forall t1 t2, In (t1,t2) ll -> size t1 = size t2) ->
       list_size size l1 = list_size size l2.
Proof.
intro size; induction ll as [ | [a b] ll ]; intros l1 l2 P1 P2 E_ll. 
rewrite (list_permut_nil P1); rewrite (list_permut_nil P2); trivial.
rewrite (list_permut_size size P1); rewrite (list_permut_size size P2);
simpl; rewrite (E_ll a b (or_introl _ (refl_equal _))).
rewrite (IHll (map (fun st : elt * elt => fst st) ll)
               (map (fun st : elt * elt => snd st) ll)); auto.
intros; apply E_ll; right; trivial.
Qed.

Add Morphism (fun size => list_size (A:=elt) size) : list_size_morph.
Proof.
apply list_permut_size.
Qed.

(** Permutation is compatible with map. *)
Lemma list_permut_map :
  forall f l1 l2, list_permut l1 l2 -> list_permut (map f l1) (map f l2).
Proof.
intros f l1; induction l1 as [ | a l1 ]; intros l2 P.
rewrite (list_permut_nil (list_permut_sym P)); apply list_permut_refl.
destruct (In_split _ _ (cons_permut_in P)) as [l2' [l2'' H]]; subst l2.
setoid_rewrite <- (permut_add_inside a nil l1) in P; simpl in P.
simpl; rewrite (IHl1 (l2' ++ l2'') P); do 2 rewrite map_app; simpl.
setoid_rewrite <- permut_cons_inside; auto.
Qed.

Add Morphism (fun l => fun f : elt -> elt => map f l) : map_morph.
Proof.
intros; apply list_permut_map; trivial.
Qed.

(** ** Permutation for short lists. *)

Lemma list_permut_length_1:
 forall a b, list_permut (a :: nil) (b :: nil)  -> a = b.
Proof.
intros a b; unfold list_permut, meq; intro P;
generalize (P a); clear P; simpl.
destruct (eq_elt_dec a a) as [ _ | a_diff_a]; [idtac | absurd (a = a); trivial].
destruct (eq_elt_dec b a) as [b_eq_a | b_diff_a]; [subst; trivial | simpl; intro; discriminate].
Qed.

Lemma list_permut_length_2 :
 forall a1 b1 a2 b2, list_permut (a1 :: b1 :: nil) (a2 :: b2 :: nil) ->
 (a1=a2 /\ b1=b2) \/ (a1=b2 /\ a2=b1).
Proof.
intros a1 b1 a2 b2 P; destruct (cons_permut_in P) as [ a2_eq_a1 | [a2_eq_b1 | a2_in_nil]].
subst a2; left; split; 
[ idtac 
| apply list_permut_length_1; setoid_rewrite (permut_cons a1)]; trivial.
subst b2; right; split;
[ idtac
| apply sym_eq; apply list_permut_length_1;
  setoid_rewrite (permut_cons_inside a1 (b1 :: nil) (a2 :: nil) nil); simpl]; trivial.
contradiction.
Qed.

(** ** Link with AC syntactic decomposition.*)
Lemma ac_syntactic_aux :
 forall (l1 l2 l3 l4 : list elt),
 list_permut (l1 ++ l2) (l3 ++ l4) ->
 (exists u1, exists u2, exists u3, exists u4, 
 list_permut l1 (u1 ++ u2) /\
 list_permut l2 (u3 ++ u4) /\
 list_permut l3 (u1 ++ u3) /\
 list_permut l4 (u2 ++ u4)).
Proof.
induction l1 as [ | a1 l1]; intros l2 l3 l4 P.
exists (nil : list elt); exists (nil : list elt); exists l3; exists l4; simpl; intuition. 

simpl in P;
destruct (in_app_or l3 l4 a1 (cons_permut_in P)) as [a1_in_l3 | a1_in_l4].
destruct (In_split _ _ a1_in_l3) as [l3' [l3'' H]]; subst l3.
rewrite app_ass in P; rewrite <- app_comm_cons in P;
setoid_rewrite <- permut_cons_inside in P;
rewrite <- app_ass in P.
destruct (IHl1 l2 (l3' ++ l3'') l4 P) as [u1 [u2 [u3 [u4 [P1 [P2 [P3 P4]]]]]]];
exists (a1 :: u1); exists u2; exists u3; exists u4; intuition; simpl; trivial.
setoid_rewrite <- P1; auto.
apply list_permut_sym; setoid_rewrite <- permut_cons_inside; auto.

destruct (In_split _ _ a1_in_l4) as [l4' [l4'' H]]; subst l4.
rewrite <- app_ass in P; 
setoid_rewrite <- permut_cons_inside in P;
rewrite app_ass in P;
destruct (IHl1 l2 l3 (l4' ++ l4'') P) as [u1 [u2 [u3 [u4 [P1 [P2 [P3 P4]]]]]]];
exists u1; exists (a1 :: u2); exists u3; exists u4; intuition; simpl; trivial.
setoid_rewrite <- permut_cons_inside; trivial.
apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside;
apply list_permut_sym; trivial.
Qed.

Lemma ac_syntactic :
 forall (l1 l2 l3 l4 : list elt),
 list_permut (l2 ++ l1) (l4 ++ l3) ->
 (exists u1, exists u2, exists u3, exists u4, 
 list_permut l1 (u1 ++ u2) /\
 list_permut l2 (u3 ++ u4) /\
 list_permut l3 (u1 ++ u3) /\
 list_permut l4 (u2 ++ u4)).
Proof.
intros l1 l2 l3 l4 P; apply ac_syntactic_aux.
apply list_permut_trans with (l2 ++ l1).
apply list_permut_app_app.
apply list_permut_trans with (l4 ++ l3); trivial.
apply list_permut_app_app.
Qed.

Function split_list 
  (l : list elt) (t : elt) {struct l} : list elt * list elt :=
  match l with
  | nil => (nil, nil)
  | a :: l' =>
      if (eq_elt_dec t a : sumbool _ _)
      then (nil, l')
      else match split_list l' t with (l1,l2) => (a :: l1, l2) end
  end.

Lemma list_permut_dec : forall (l1 l2 : list elt), {list_permut l1 l2}+{~list_permut l1 l2}.
Proof.
intro l1; induction l1 as [ | e1 l1].
destruct l2 as [ | e2 l2].
left; apply list_permut_refl.
right; intro P; assert (H := list_permut_length P); discriminate.
intro l2; destruct (In_dec eq_elt_dec e1 l2) as [e1_in_l2 | e1_not_in_l2].
assert (H:= @f_equal _ _ (fun l => split_list l e1) l2); simpl in H.
destruct (split_list l2 e1) as [l2' l2''].
generalize (sym_eq (H _ (refl_equal _))); clear H; intro H.
assert (H' : l2 = l2' ++ e1 :: l2'').
generalize l2' l2'' H; clear l2' l2'' H; functional induction (split_list l2 e1)
    as [ | l a a' l' l_eq_a'_l' a_eq_a' _ | l a a' l' l_eq_a'_l' a_diff_a' _ IH ll1 ll2];
intros l2' l2'' H'.
contradiction.
injection H'; intros; subst; simpl; trivial.
injection H'; intros; simpl; subst; simpl; 
apply (f_equal (fun l => a' :: l)); apply IH; trivial.
destruct e1_in_l2 as [a_eq_a' | a_in_l']; [absurd (a = a') | idtac]; subst; trivial.
destruct (IHl1 (l2' ++ l2'')) as [P | nP].
left; subst; setoid_rewrite <- permut_cons_inside; trivial.
right; intro P; apply nP; setoid_rewrite (permut_cons_inside e1); subst; trivial.
right; intro P; apply e1_not_in_l2; setoid_rewrite <- (in_permut_in e1 P); left; trivial.
Qed.

Lemma permut_filter :
 forall P (P_dec : forall a, {P a}+{~P a}) (l : list elt),
  list_permut l ((filter (fun a => if P_dec a then true else false) l)++
                         (filter (fun a => if P_dec a then false else true) l)).
Proof.
intros P P_dec; induction l as [ | a l]; simpl; auto.
destruct (P_dec a) as [Pa | not_Pa].
rewrite <- app_comm_cons; setoid_rewrite <- IHl; auto. 
setoid_rewrite <- permut_cons_inside; trivial.
Qed.

Lemma permut_insert_2 :
forall (l1 l2 r1 r2 r3 : list elt) (a b b' : elt), 
  list_permut (l1 ++ l2) (r1 ++ b :: r2 ++ b' :: r3) ->
 exists r1', exists r2', exists r3', 
 list_permut (l1 ++ a :: l2)  (r1' ++ b :: r2' ++ b' :: r3').
Proof.
intros l1; induction l1 as [ | a1 l1].
simpl; intros l2 r1 r2 r3 a b b' H.
exists (a :: r1); exists r2; exists r3; simpl; setoid_rewrite H; auto.
intros l2 r1 r2 r3 a b b' P;
assert (P' : list_permut ((a1 :: l1) ++ l2) (l1 ++ (a1 :: l2))).
simpl; setoid_rewrite <- permut_cons_inside; auto.
setoid_rewrite P' in P;
destruct (IHl1 (a1 :: l2) r1 r2 r3 a b b' P) as [r1' [r2' [r3' P'']]];
exists r1'; exists r2'; exists r3'; setoid_rewrite <- P'';
simpl; setoid_rewrite <- (permut_add_inside a (a1 :: l1) l2 l1 (a1 :: l2)); auto.
Qed.

Lemma permut_list_exists :
  forall f l l', list_permut l l' -> list_exists f l = list_exists f l'.
Proof.
intros f l; induction l as [ | a l]; simpl; intros l' P.
rewrite (list_permut_nil (list_permut_sym P)); simpl; trivial.
assert (a_in_l' : In a l').
setoid_rewrite <- P; left; trivial.
destruct (In_split _ _ a_in_l') as [k' [k'' H]]; subst l'.
rewrite list_exists_app; simpl.
destruct (f a); simpl.
destruct (list_exists f k'); simpl; trivial.
rewrite <- list_exists_app; apply IHl;
setoid_rewrite (permut_cons_inside a); trivial.
Qed.

Lemma permut_list_forall_exists :
  forall f l1 l1' l2 l2', list_permut l1 l1' -> list_permut l2 l2' -> 
  list_forall (fun t1 => list_exists (fun t2 => f t1 t2) l2) l1 = 
  list_forall (fun t1 => list_exists (fun t2 => f t1 t2) l2') l1'. 
Proof.
intros f l; induction l as [ | a l]; simpl; intros l' l2 l2' P P2.
rewrite (list_permut_nil (list_permut_sym P)); simpl; trivial.
assert (a_in_l' : In a l').
setoid_rewrite <- P; left; trivial.
destruct (In_split _ _ a_in_l') as [k' [k'' H]]; subst l'.
rewrite list_forall_app; simpl.
rewrite <- (permut_list_exists (fun t2 : elt => f a t2) P2).
destruct (list_exists (fun t2 : elt => f a t2) l2); simpl.
rewrite <- list_forall_app; apply IHl; trivial;
setoid_rewrite (permut_cons_inside a); trivial.
destruct (list_forall (fun t1 : elt => list_exists (fun t2 : elt => f t1 t2) l2') k'); 
simpl; trivial.
Qed.

Lemma remove_is_sound :
  forall x l, match remove eq_elt_dec x l with
                | None => ~In x l
                | Some l' => list_permut l (x :: l')
                end.
Proof.
intros x l; generalize (in_remove eq_elt_dec x l);
destruct (remove eq_elt_dec x l) as [ l' | ]; trivial.
intros [x_in_l [l1 [l2 [H1 H2]]]]; injection H2; intros; subst;
apply list_permut_sym; setoid_rewrite <- permut_cons_inside; auto.
Qed.


Function remove_equiv  (l1 l2 : list elt) {struct l1} : (list elt) * (list elt) := 
  match l1 with 
    | nil => (l1,  l2)
    | x1 :: l1' =>
        match remove DS.eq_A_dec x1 l2 with
        | Some l2' => remove_equiv l1' l2'
        | None =>
              match remove_equiv l1' l2 with
              | (l1'', l2'') => ((x1 :: l1'') , l2'')
              end
          end
     end.

Lemma remove_equiv_is_sound: 
  forall l1 l2, match remove_equiv l1 l2 with
  | (l1', l2') => exists l, list_permut l1 (l ++ l1') /\ list_permut l2 (l ++ l2') /\
                        (forall x, In x l1' -> In x l2' -> False)
  end.
Proof.
intros l1 l2; 
functional induction (remove_equiv l1 l2) 
    as [ (* Case 1 *)
         | (* Case 2 *) H1 l2 e1 l1 H2 l2' H H' 
         | (* Case 3 *) H1 l2 e1 l1 H2 H H' l1'' l2'' H''].
(* Case 1 *)
exists (@nil elt); simpl; intuition.
(* Case 2 *)
destruct (remove_equiv l1 l2') as [l1'' l2'']; destruct H' as [l [P1 [P2 H']]].
exists (e1 :: l); simpl; split.
setoid_rewrite P1; auto.
split; trivial.
setoid_rewrite <- P2;
assert (H'' := remove_is_sound e1 l2); unfold eq_elt_dec, elt in *; 
rewrite H in H''; trivial.
(* Case 3 *)
destruct (remove_equiv l1 l2) as [l1' l2']; injection H''; clear H''; intros; subst l1'' l2'';
destruct H' as [l [P1 [P2 H']]]; exists l; split.
apply (proj1 (permut_cons_inside e1 l1 l l1')); trivial.
split; trivial.
simpl; intros x [x_eq_e1 | x_in_l1']  x_in_l2.
assert (H'' := remove_is_sound e1 l2); unfold eq_elt_dec, elt in *; 
rewrite H in H''; subst x; absurd (In e1 l2); trivial.
setoid_rewrite (in_permut_in e1 P2); apply in_or_app; right; trivial.
apply (H' x); trivial.
Qed.

Lemma split_equiv_aux:
  forall (R : relation elt), equivalence _ R -> forall ll s s',
   (forall t1 t2, In (t1,t2) ll -> R t1 t2) ->
   In s (map (fun st => fst st) ll) ->
   In s' (map (fun st => snd st) ll) -> 
   R s s' ->
   exists ll', (forall t1 t2, In (t1, t2) ll' -> R t1 t2) /\
                  list_permut  (map (fun st => fst st) ll) (s :: map (fun st => fst st) ll') /\
                  list_permut (map (fun st => snd st) ll) (s' :: map (fun st => snd st) ll').
Proof.
intros R equiv_R ll; induction ll as [ | [t t'] ll]; intros s s' E_ll s_in_ll1 s'_in_ll2 s_eq_s'.
simpl in s_in_ll1; contradiction.
simpl in s_in_ll1; destruct s_in_ll1 as [s_eq_t | s_in_ll1];
simpl in s'_in_ll2; destruct s'_in_ll2 as [s'_eq_t' | s'_in_ll2].
(* s = t, s' = t' *)
subst t t'; exists ll; simpl; repeat split; auto; intros; apply E_ll; right; trivial.
(* s = t, s' in ll2 *)
subst t; destruct (equiv_in_list_2 _ R s' ll) as [t [ll1 [ll2 [t_eq_s' H]]]]; auto.
intros; apply E_ll; right; trivial.
destruct (IHll t s') as [ll' [E_ll' [Q Q']]]; auto.
intros; apply E_ll; right; trivial.
subst ll; rewrite map_app; apply in_or_app; right; left; trivial.
exists ((t,t') :: ll'); repeat split.
intros t1 t2 [t1t2_eq_st | t1t2_in_ll'].
injection t1t2_eq_st; intros; subst t1 t2.
apply (equiv_trans _ _ equiv_R) with s'; trivial.
apply (equiv_trans _ _ equiv_R) with s.
apply (equiv_sym _ _ equiv_R); trivial.
apply E_ll; left; trivial.
apply E_ll'; trivial.
simpl; setoid_rewrite <- permut_cons; setoid_rewrite Q; auto.
simpl; setoid_rewrite <- (permut_cons_inside t' (map (fun st => snd st) ll) (s' :: nil)
                                                           (map (fun st => snd st) ll'));
simpl; trivial.
(* s in ll1, s' = t' *)
subst t'; destruct (equiv_in_list _ R s ll) as [t' [ll1 [ll2 [s_eq_t' H]]]]; auto.
intros; apply E_ll; right; trivial.
destruct (IHll s t') as [ll' [E_ll' [Q Q']]]; auto.
intros; apply E_ll; right; trivial.
subst ll; rewrite map_app; apply in_or_app; right; left; trivial.
exists ((t,t') :: ll'); repeat split.
intros t1 t2 [t1t2_eq_st | t1t2_in_ll'].
injection t1t2_eq_st; intros; subst t1 t2; 
apply (equiv_trans _ _ equiv_R) with s'.
apply E_ll; left; trivial.
apply (equiv_trans _ _ equiv_R) with s; trivial.
apply (equiv_sym _ _ equiv_R); trivial.
apply E_ll'; trivial.
simpl; setoid_rewrite <- (permut_cons_inside t (map (fun st => fst st) ll) (s :: nil)
                                                           (map (fun st => fst st) ll'));
simpl; trivial.
simpl; setoid_rewrite <- permut_cons; setoid_rewrite Q'; auto.
(* s in ll1, s' in ll2 *)
destruct (IHll s s') as [ll' [E_ll' [Q Q']]]; auto.
intros; apply E_ll; right; trivial.
exists ((t,t') :: ll'); repeat split.
intros t1 t2 [t1t2_eq_tt' | t1t2_in_ll'].
injection t1t2_eq_tt'; intros; subst t1 t2; apply E_ll; left; trivial.
apply E_ll'; trivial.
simpl; setoid_rewrite <- (permut_cons_inside t (map (fun st => fst st) ll) (s :: nil)
                                                           (map (fun st => fst st) ll'));
simpl; trivial.
simpl; setoid_rewrite <- (permut_cons_inside t' (map (fun st => snd st) ll) (s' :: nil)
                                                           (map (fun st => snd st) ll'));
simpl; trivial.
Qed.

Lemma split_equiv :
  forall (R : relation elt), equivalence _ R -> 
  forall ll l1 l2, (forall t1 t2, In (t1,t2) ll -> R t1 t2) ->
        list_permut (map (fun st => fst st) ll) (l1 ++ l2) ->
       exists ll1, exists ll2,
       (forall t1 t2, In (t1,t2) ll1 -> R t1 t2) /\
       (forall t1 t2, In (t1,t2) ll2 -> R t1 t2) /\
       list_permut l1 (map (fun st => fst st) ll1) /\
       list_permut l2 (map (fun st => fst st) ll2) /\
       list_permut (map (fun st => snd st) ll) 
                           (map (fun st => snd st)  (ll1 ++ ll2)).
Proof. 
intros R equiv_R ll; induction ll as [ | [s t] ll]. 
intros l1 l2 _ P; simpl in P; 
assert (l1_eq_nil : l1 = nil).
assert (L := list_permut_length P); rewrite length_app in L.
destruct l1 as [ | s1 l1]; trivial; simpl in L; discriminate.
subst l1; assert (l2_eq_nil : l2 = nil).
assert (L := list_permut_length P); rewrite length_app in L.
destruct l2 as [ | s2 l2]; trivial; simpl in L; discriminate.
subst l2; clear P;
exists (@nil (elt * elt)); exists (@nil (elt * elt)); repeat split.
intros; contradiction.
intros; contradiction.
intros l1 l2 E_st_ll P; simpl in P.
assert (s_in_l1l2 : In s (l1 ++ l2)).
setoid_rewrite <- (in_permut_in s P); left; trivial.
destruct (in_app_or _ _ _ s_in_l1l2) as [s_in_l1 | s_in_l2].
destruct (In_split _ _ s_in_l1) as [l1' [l1'' H]]; subst l1.
destruct (IHll (l1' ++ l1'') l2) as [ll1 [ll2 [E_ll1 [E_ll2 [P1 [P2 P']]]]]].
intros; apply E_st_ll; right; trivial.
rewrite <- ass_app; setoid_rewrite (permut_cons_inside s);
rewrite app_comm_cons; rewrite ass_app; trivial.
exists ((s,t) :: ll1); exists ll2; repeat split; trivial.
intros t1 t2 [t1t2_eq_st | t1t2_in_ll1].
injection t1t2_eq_st; intros; subst t1 t2; apply E_st_ll; left; trivial.
apply E_ll1; trivial.
simpl; apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
apply list_permut_sym; trivial.
simpl; setoid_rewrite <- permut_cons; trivial.
destruct (In_split _ _ s_in_l2) as [l2' [l2'' H]]; subst l2.
destruct (IHll l1 (l2' ++ l2'')) as [ll1 [ll2 [E_ll1 [E_ll2 [P1 [P2 P']]]]]].
intros; apply E_st_ll; right; trivial.
rewrite ass_app; setoid_rewrite (permut_cons_inside s);
rewrite <- ass_app; trivial.
exists ll1; exists ((s,t) :: ll2); repeat split; trivial.
intros t1 t2 [t1t2_eq_st | t1t2_in_ll2].
injection t1t2_eq_st; intros; subst t1 t2; apply E_st_ll; left; trivial.
apply E_ll2; trivial.
simpl; apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
apply list_permut_sym; trivial.
rewrite map_app; simpl; setoid_rewrite <- permut_cons_inside;
rewrite <- map_app; trivial.
Qed.

Lemma split_equiv_2 :
  forall (R : relation elt), equivalence _ R -> 
  forall ll l1 l2, (forall t1 t2, In (t1,t2) ll -> R t1 t2) ->
        list_permut (map (fun st => snd st) ll) (l1 ++ l2) ->
       exists ll1, exists ll2,
       (forall t1 t2, In (t1,t2) ll1 -> R t1 t2) /\
       (forall t1 t2, In (t1,t2) ll2 -> R t1 t2) /\
       list_permut l1 (map (fun st => snd st) ll1) /\
       list_permut l2 (map (fun st => snd st) ll2) /\
       list_permut (map (fun st => fst st) ll) 
                           (map (fun st => fst st)  (ll1 ++ ll2)).
Proof.
intros R equiv_R ll l1 l2 E_ll P;
destruct (equiv_swap _ R ll E_ll) as [ll' [E_ll' [H1 [H2 H3]]]].
rewrite H2 in P.
destruct (split_equiv equiv_R ll' l1 l2) 
  as [ll1 [ll2 [E_ll1 [E_ll2 [H1' [H2' H3']]]]]]; trivial.
intros; apply (equiv_sym _ _ equiv_R); apply E_ll; setoid_rewrite H3; trivial.
destruct (equiv_swap _ R ll1 E_ll1) as [ll1' [E_ll1' [H1'' [H2'' H3'']]]].
destruct (equiv_swap _ R ll2 E_ll2) as [ll2' [E_ll2' [H1''' [H2''' H3''']]]].
exists ll1'; exists ll2'; repeat split; trivial.
intros; apply (equiv_sym _ _ equiv_R); apply E_ll1; setoid_rewrite H3''; trivial.
intros; apply (equiv_sym _ _ equiv_R); apply E_ll2; setoid_rewrite H3'''; trivial.
rewrite <- H1''; trivial.
rewrite <- H1'''; trivial.
rewrite H1; rewrite map_app; rewrite <- H2''; rewrite <- H2'''; 
rewrite <- map_app; trivial.
Qed.

Lemma trans_equiv_list :
  forall (R : relation elt), forall ll ll',
  (forall e1, In e1 (map (fun st => fst st) ll) ->
        forall e2 e3, R e1 e2 -> R e2 e3 -> R e1 e3) ->
  (forall e1 e2, In (e1,e2) ll -> R e1 e2) -> 
  (forall e1 e2, In (e1,e2) ll' -> R e1 e2) -> 
  list_permut (map (fun st => snd st) ll) (map (fun st => fst st) ll') -> 
     exists ll'', (forall e1 e2, In (e1,e2) ll'' -> R e1 e2) /\
             list_permut (map (fun st => fst st) ll)
                                 (map (fun st => fst st) ll'') /\
             list_permut (map (fun st => snd st) ll') 
                                 (map (fun st => snd st) ll'').
Proof.
intros R ll; induction ll as [ | [s t] ll].
intros [ | [s' t'] ll'] trans_R E_ll E_ll' P.
exists (nil : list (elt * elt)); intuition.
generalize (list_permut_length P); intro L; inversion L.
intros ll' trans_R E_st_ll E_ll' P.
assert (t_in_ll1' := (proj1 (in_permut_in t P)) (or_introl _ (refl_equal _))).
destruct (equiv_in_list _ R _ _ E_ll' t_in_ll1') as [t' [ll1' [ll2' [t_eq_t' H]]]]; subst  ll'.
assert (E_ll : forall t1 t2, In (t1, t2) ll -> R t1 t2).
intros; apply E_st_ll; right; trivial.
assert (IH_ll : forall t, In t (map (fun st => fst st) ll) ->
        forall y z, R t y -> R y z -> R t z).
intros u u_in_ll1; apply trans_R; right; trivial.
destruct (IHll (ll1' ++ ll2') IH_ll E_ll) as [ll'' [E_ll'' [P1'' P2'']]].
intros; apply E_ll'; apply in_insert; trivial.
rewrite map_app in P; simpl in P; rewrite <- permut_cons_inside in P;
rewrite <- map_app in P; trivial.
exists ((s,t') :: ll''); repeat split.
simpl; intros t2 t3 [t2t3_eq_st' | t2t3_in_ll''].
injection t2t3_eq_st'; intros; subst t2 t3; apply trans_R with t.
simpl; left; trivial.
apply E_st_ll; left; trivial.
apply E_ll'; apply in_or_app; right; left; trivial.
apply E_ll''; trivial.
simpl; setoid_rewrite <- permut_cons; trivial.
rewrite map_app; simpl; apply list_permut_sym;
setoid_rewrite <- permut_cons_inside; rewrite <- map_app;
apply list_permut_sym; trivial.
Qed.
 

End Make.

(* With section instead of module, and then polymorph use 
Section LP.
Variable elt : Set.
Parameter eq_elt_dec : forall t1 t2 : elt, {t1 = t2} + {t1 <> t2}.
...
End LP.

Add Relation list list_permut 
reflexivity proved by list_permut_refl
symmetry proved by list_permut_sym
transitivity proved by list_permut_trans as LP_poly.

Add Morphism In with signature  eq ==> list_permut ==> iff as in_morph_poly.
Proof.
intros; split; intros.
apply in_permut_in with x1; trivial.
apply in_permut_in with x2; trivial; apply list_permut_sym; trivial.
Qed.

Goal forall (a b:nat), list_permut (a::b::nil) (b::a::nil) -> 
In a (b :: a :: nil).
intros a b P.
setoid_rewrite <- P.
left; trivial.
Qed.
*)

(*
Extract Constant eq_element_dec => eq.
Extract Constant o_element_dec => le.
Extract Constant element => int.
Extraction split_list.
Extraction partition.
Extraction NoInline partition.
Extraction quicksort.
*)

