Add LoadPath "basis". 

Set Implicit Arguments.

Require Import Relations.
Require Import List.
Require Import more_list.
Require Import Multiset.
Require Import list_permut.
Require Import Arith.
Require Import ordered_set.
Require Import Recdef.

(** ** Definition of quicksort over lists. *)
Module Make (DS1 : decidable_set.S)
(DOS1 : ordered_set.S with Definition A:= DS1.A).

Module LP := list_permut.Make (DS1).

Import LP.
Definition o_elt := DOS1.o_A.
Definition o_elt_proof := DOS1.o_A_proof.
Definition o_elt_dec := DOS1.o_A_dec.
Definition o_elt_total := DOS1.o_A_total.

(** *** Decomposition of the list wrt a pivot. *)
Function partition (pivot : elt) (l : list elt) {struct l} : (list elt) * (list elt) :=
  match l with 
  | nil => (nil, nil)
  | a :: l' => 
	match partition pivot l' with (l1,l2) =>
	 if ((o_elt_dec a pivot) : sumbool _ _) then (a :: l1, l2) else (l1, a :: l2)
        end
end.

Function quicksort (l : list elt) { measure (@length elt) l} : list elt :=
   match l with
	| nil => nil
	| h :: tl => 
             match partition h tl with (l1, l2) =>
             quicksort l1 ++ h :: quicksort l2
        end
    end. 
Proof. 
(* 1 Second recursive call *)
intros e_l e l e_l_eq_e_l; subst e_l;
functional induction (partition e l) 
   as [ | H1 e' l' H2 IH l1 l2 rec_res e'_lt_e _ | H1 e' l' H2 IH l1 l2 rec_res H3 H4];
intros ll1 ll2 Hpart; injection Hpart; intros H H'; subst.
(* 1.1 l = nil *)
simpl; auto with arith.
(* 1.2 l = _ :: _ *)
apply lt_trans with (length (e :: l')); [apply (IH _ _ rec_res) | auto with arith].
simpl; apply lt_n_S; apply (IH _ _ rec_res).

(* 2 First recursive call *)
intros e_l e l e_l_eq_e_l; subst e_l; 
functional induction (partition e l)
   as [ | H1 e' l' H2 IH l1 l2 rec_res e'_lt_e _ | H1 e' l' H2 IH l1 l2 rec_res H3 H4];
intros ll1 ll2 Hpart; injection Hpart; intros H H'; subst.
(* 2.1 l = nil *)
simpl; auto with arith.
(* 2.2 l = _ :: _ *)
simpl; apply lt_n_S; apply (IH _ _ rec_res).
apply lt_trans with (length (e :: l')); [apply (IH _ _ rec_res) | auto with arith].
Qed.

(** *** The result of quicksort is a permutation of the original list. *)
Lemma partition_list_permut1 :
 forall e l, let (l1,l2) := partition e l in list_permut (l1 ++ l2) l.
Proof.
intros e l; functional induction (partition e l)
   as [ | H1 e' l' H2 IH l1 l2 rec_res e'_lt_e _ | H1 e' l' H2 IH l1 l2 rec_res H3].
 (* l = nil *)
auto.
(* l = a :: l' /\ o_A a e *)
simpl app; rewrite rec_res in IH; setoid_rewrite IH; auto.
 (* l = a::l' /\ ~o_A a e *)
simpl app; rewrite rec_res in IH;
apply list_permut_sym; setoid_rewrite <- permut_cons_inside; auto.
Qed.

Theorem quick_permut : forall l, list_permut l (quicksort l).
Proof.
intros l; functional induction (quicksort l)
   as [ | l a' l' l_eq_a'_l' l1 l2 H S1 S2]; auto.
(* 2.1 l = nil *)
setoid_rewrite <- permut_cons_inside.
assert (P := partition_list_permut1 a' l'); rewrite H in P.
setoid_rewrite <- P; setoid_rewrite <- S1; setoid_rewrite <- S2; auto.
Qed.

Theorem quick_permut_bis :
 forall l1 l2, list_permut l1 l2 -> list_permut (quicksort l1) l2.
Proof.
intros l1 l2 P; apply list_permut_trans with l1; trivial;
apply list_permut_sym; apply quick_permut.
Qed.

(** *** Definition of a sorted list. *)
Inductive is_sorted : list elt -> Prop :=
  | is_sorted_nil : is_sorted nil
  | is_sorted_cons1 : forall t : elt, is_sorted (t :: nil)
  | is_sorted_cons2 :
      forall (t1 t2:elt) (l:list elt),
        o_elt t1 t2 -> is_sorted (t2 :: l) -> is_sorted (t1 :: t2 :: l).

Hint Constructors is_sorted.

(** *** The result of quicksort is a sorted list. *)
Lemma sorted_tl_sorted :
  forall e l, is_sorted (e :: l) -> is_sorted l.
Proof.
intros e l S; inversion S; auto.
Qed.

Lemma quick_sorted_aux1 :
  forall l1 l2 e, is_sorted l1 -> is_sorted l2 ->
  (forall a, In a l1 -> o_elt a e) ->
  (forall a, In a l2 -> o_elt e a) -> 
  is_sorted (l1 ++ e :: l2).
Proof.
induction l1 as [ | e1 l1 ]; intros l2 e S1 S2 L1 L2.
simpl; destruct l2 as [ | e2 l2]; intros; 
[apply is_sorted_cons1 | apply is_sorted_cons2 ]; trivial;
apply (L2 e2); simpl; left; trivial.
destruct l1 as [ | e1' l1] ; simpl; intros; apply is_sorted_cons2.
apply L1; left; trivial.
simpl in IHl1; apply IHl1; trivial; contradiction.
inversion S1; trivial.
rewrite app_comm_cons; apply IHl1; trivial.
inversion S1; trivial.
intros; apply L1; trivial; right; trivial.
Qed.

Lemma quick_sorted_aux3 :
  forall l e, let (l1,_) := partition e l in forall a, In a l1 -> o_elt a e.
Proof.
intros l e; 
functional induction (partition e l) 
    as [ | l a' l' l_eq_a'_l' IH l1' l2' H a'_lt_e _ | l a' l' l_eq_a'_l' IH l1' l2' H _H _ ].
contradiction.
intros a [a_eq_a' | a_in_l1']; [subst | rewrite H in IH; apply IH]; trivial.
intros; rewrite H in IH; apply IH; trivial.
Qed.

Lemma quick_sorted_aux4 :
  forall l e, let (_,l2) := partition e l in forall a, In a l2 -> o_elt e a.
Proof.
intros l e; 
functional induction (partition e l) 
    as [ | l a' l' l_eq_a'_l' IH l1' l2' H a'_lt_e _ | l a' l' l_eq_a'_l' IH l1' l2' H a'_ge_e _ ].
contradiction.
intros; rewrite H in IH; apply IH; trivial.
intros a [a_eq_a' | a_in_l1']; [subst | rewrite H in IH; apply IH]; trivial.
destruct (o_elt_total e a); trivial; absurd (o_elt a e); trivial.
Qed.

Theorem quick_sorted : forall l, is_sorted (quicksort l).
Proof.
intros l; 
functional induction (quicksort l)
   as [ | l a' l' l_eq_a'_l' l1 l2 H S1 S2].
auto.
apply quick_sorted_aux1; trivial.
intros a a_in_ql1; 
assert (H1 := quick_sorted_aux3 l' a'); rewrite H in H1;
apply H1; setoid_rewrite (in_permut_in a (quick_permut l1)); trivial.
intros a a_in_ql2; 
assert (H2 := quick_sorted_aux4 l' a'); rewrite H in H2;
apply H2; setoid_rewrite (in_permut_in a (quick_permut l2)); trivial.
Qed.

(** ***  There is a unique sorted list equivalent up to a permutation to a given list.*)
Lemma sorted_cons_min : 
 forall l e, is_sorted (e :: l) -> (forall e', In e' (e :: l) -> o_elt e e').
Proof.
intros l e S e' In_e'; elim In_e'; clear In_e'; intro In_e'; subst.
generalize o_elt_proof; intro O; inversion O as [R _ _]; apply R.
generalize e S e' In_e'; clear e S e' In_e'; induction l as [ | a l ];
intros e S e' In_e'; inversion S; elim In_e'; clear In_e'; intro In_e'; 
subst; trivial.
generalize o_elt_proof; intro O; inversion O as [ _ T _ ]; 
apply T with a; trivial; apply IHl; trivial.
Qed.

Theorem sort_is_unique :
  forall l1 l2, is_sorted l1 -> is_sorted l2 -> list_permut l1 l2 -> l1 = l2.
Proof.
induction l1 as [ | e1 l1 ]; intros l2 S1 S2 P.
apply sym_eq; apply list_permut_nil; auto.
destruct l2 as [ | e2 l2 ].
apply (list_permut_nil P).
assert (e1_eq_e2 : e1 = e2).
assert (O:= o_elt_proof); inversion O as [ _ _ AS ]; apply AS; clear O AS.
apply (sorted_cons_min S1); setoid_rewrite P; left; trivial.
apply (sorted_cons_min S2); setoid_rewrite <- P; left; trivial.
subst; rewrite (IHl1 l2); trivial.
apply sorted_tl_sorted with e2; trivial.
apply sorted_tl_sorted with e2; trivial.
setoid_rewrite (permut_cons e2 l1 l2); trivial.
Qed.

(** Some usefull properties on the result of quicksort. *)
Lemma length_quicksort : forall l, length (quicksort l) = length l.
Proof.
intro l; apply list_permut_length;
apply list_permut_sym; apply quick_permut.
Qed.

Lemma in_quicksort_cons : forall a l, In a (quicksort (a :: l)).
Proof.
intros a l; setoid_rewrite <- (in_permut_in a (quick_permut (a :: l))); left; trivial.
Qed.

(** What happens when a sorted list is removed from another one.*)
Function remove_list (la l : list elt) {struct l} : option (list elt) :=
  match la with
  | nil => Some l
  | a :: la' => 
	match l with 
	| nil => None
	| b :: l' => 
	   if (eq_elt_dec a b : sumbool _ _)
	   then remove_list la' l'
	   else 
	     match remove_list la l' with
	     | None => None
	     | Some rmv => Some (b :: rmv)
	     end
        end
  end.

Lemma in_remove_list :
  forall l la, is_sorted la -> is_sorted l -> 
  match remove_list la l with
  | None => forall rmv, ~ (list_permut (la ++ rmv) l)
  | Some rmv => list_permut (la ++ rmv) l
  end.
Proof.
intros l la Sa S; 
functional induction (remove_list la l) 
   as [ (* Case 1 *)
        | (* Case 2 *) la l a' la' H1 H2 
        | (* Case 3 *) la l a' la' H1 e' l' H2 H3 _ IH 
        | (* Case 4 *) la l a' la' H1 e' l' H2 a'_diff_e' _H IH H
        | (* Case 5 *) la l a' la' H1 e' l' H2 a'_diff_e' _H IH rmv H].
(* Case 1 *)
auto.
(* Case 2 *)
intros rmv P; assert (H := list_permut_length P); discriminate.
(* Case 3 *)
destruct (remove_list la' l') as [rmv | ].
simpl; setoid_rewrite <- permut_cons; 
apply IH; apply sorted_tl_sorted with a'; trivial.
intros rmv P; refine (IH _ _ rmv _);
[ apply sorted_tl_sorted with a'
| apply sorted_tl_sorted with a'
| setoid_rewrite (permut_cons a'); simpl in P ]; trivial.
(* Case 4 *)
rewrite H in IH.
intros rmv P; assert (H' : In e' ((a' :: la') ++rmv)).
setoid_rewrite P; left; trivial.
simpl in H'; destruct H' as [e'_eq_a' | H'].
apply a'_diff_e'; trivial.
destruct (in_app_or _ _ _ H') as [e'_in_la' | e'_in_rmv].
apply a'_diff_e';
assert (O := o_elt_proof); inversion_clear O as [ _ _ AS]; apply AS.
apply sorted_cons_min with la'; [apply Sa | right; trivial].
apply sorted_cons_min with l'; [apply S | setoid_rewrite <- P; left; trivial].
destruct (In_split _ _ e'_in_rmv) as [l1 [l2 H'']]; subst rmv.
refine (IH _ _ (l1 ++ l2) _).
trivial.
apply sorted_tl_sorted with e'; trivial.
setoid_rewrite (permut_cons e'); setoid_rewrite <- P;
do 2 rewrite <- app_ass; setoid_rewrite <- permut_cons_inside; auto.
(* Case 5 *)
rewrite H in IH.
apply list_permut_sym; setoid_rewrite <- permut_cons_inside; 
setoid_rewrite <- IH; subst; auto; apply sorted_tl_sorted with e'; trivial.
Qed.

End Make.

(*
(** *** The lists resulting from the partition are not longer than the original one. *)
Lemma length_fst_partition :
 forall n, forall pivot, forall l,
  length l <= n -> length (fst (partition pivot l)) < S n.
Proof.
induction n; intros pivot l; destruct l as [ | e l ]; simpl; auto with arith.
intro H; generalize (le_Sn_O (length l) H); intros; contradiction.
intro L'; assert (L : length l <= n).
apply le_S_n; trivial.
generalize (IHn pivot l L); destruct (partition pivot l);
destruct (o_elt_dec e pivot); simpl; auto with arith.
Qed.

Lemma length_snd_partition :
 forall n, forall pivot, forall l, 
  length l <= n -> length (snd (partition pivot l)) < S n.
Proof.
induction n; intros pivot l; destruct l as [ | e l ]; simpl; auto with arith.
intro H; generalize (le_Sn_O (length l) H); intros; contradiction.
intro L'; assert (L : length l <= n).
apply le_S_n; trivial.
generalize (IHn pivot l L); destruct (partition pivot l);
destruct (o_elt_dec e pivot); simpl; auto with arith.
Qed.

(** *** Definition of a step of quicksort. *)
Definition F_quick : 
  forall l : list elt, (forall y : list elt , o_length y l -> list elt) -> list elt.
Proof.
destruct l as [ | e l ].
intros; exact nil.
intro qrec;
set (pl := partition e l);
exact 
((qrec (fst pl) (length_fst_partition e l (le_n (length l)))) ++
 e :: (qrec (snd pl) (length_snd_partition e l (le_n (length l))))).
Defined.

(** *** Definition of quicksort. *)
Definition quicksort : forall (l : list elt), (list elt) :=
Fix (@well_founded_length elt) (fun l => list elt) F_quick.

(** *** A more practical equivalent definition of quicksort. *)
Lemma quicksort_unfold : forall l : list elt,
quicksort l = @F_quick l (fun y _ => quicksort y).
Proof.
intros; unfold quicksort;
apply Fix_eq with (P:= (fun _:list elt => list elt)).
unfold F_quick; destruct x; simpl; auto.
intros f g H; do 2 rewrite H; trivial.
Save.

(** *** Sorting the empty list yields the empty list. *)
Lemma quick_nil : quicksort nil = nil.
Proof.
rewrite quicksort_unfold; unfold F_quick; simpl; trivial.
Qed.

(** *** The usual equality on quicksort. *)
Lemma quick_cons_aux1 : 
  forall e l, quicksort (e :: l) = 
              let pl := partition e l in 
              (quicksort (fst pl)) ++ e :: (quicksort (snd pl)).
Proof.
intros e l; rewrite quicksort_unfold; unfold F_quick; simpl; trivial.
Qed.

Theorem quick_cons : 
 forall e l, let (l1,l2) := partition e l in 
 quicksort (e :: l) = (quicksort l1) ++ e :: (quicksort l2).
Proof.
intros; generalize (quick_cons_aux1 e l); destruct (partition e l).
trivial.
Save.

Extract Constant eq_elt_dec => eq.
Extract Constant o_elt_dec => le.
Extract Constant elt => int.
Extraction split_list.
Extraction partition.
Extraction NoInline partition.
Extraction quicksort.
*)

